unit rdbtool_MDB;

{<
  rdbtool_MDB is set functions for working with MDB database file
  by using of DAO; can be used for updating of database structure
  by application.


  Date: Dec 2019
  Author: Rosi (http://www.rosinsky.cz/delphi)
  Full functional demo
}

interface

{$INCLUDE _rosi.inc}

uses Windows, Classes, Controls, SysUtils, ActiveX, Variants,
{$IFDEF DX100P} dao2010, {$ELSE} dao2000, {$ENDIF}
  comobj;

type
  TDBVersionRec = record
    Value: integer;
    Name: string;
    Extension: string;
  end;

  TDBLocaleRec = record
    Value: integer;
    Name: string;
    Setting: string;
  end;

  TDBFieldTypeRec = record
    Value: integer;
    Name: string;
    Desc: string;
  end;

const
{$IFDEF DX100P}
  DBVersionArrSize = 4;
  DBVersionDefault =  dbVersion140;
  DBVersionArr: array [1 .. DBVersionArrSize] of TDBVersionRec =  (
  (Value: dbVersion30; Name: 'Version 3.0 (MDB)'; Extension: '.mdb'),
  (Value: dbVersion40; Name: 'Version 4.0 (MDB)'; Extension: '.mdb'),
  (Value: dbVersion120; Name: 'Version 12.0 (ACCDB)'; Extension: '.accdb'),
  (Value: dbVersion140; Name: 'Version 14.0 (ACCDB)'; Extension: '.accdb')
  );
{$ELSE}
  DBVersionArrSize = 2;
  DBVersionDefault =  dbVersion40;
  DBVersionArr: array [1 .. DBVersionArrSize] of TDBVersionRec =  (
  (Value: dbVersion30; Name: 'Version 3.0 (MDB)'; Extension: '.mdb'),
  (Value: dbVersion40; Name: 'Version 4.0 (MDB)'; Extension: '.mdb')
  );
{$ENDIF}
  DBVersionMDB = dbVersion40;

  DBLocaleArrSize = 20;
  DBLocaleDefault = dbSortGeneral;
  DBLocaleArr: array [1 .. DBLocaleArrSize] of TDBLocaleRec = (
    (Value: dbSortGeneral; Name: 'General (English, French, German ...)'; Setting: ';LANGID=0x0409;CP=1252;COUNTRY=0'),
    (Value: dbSortArabic; Name: 'Arabic'; Setting: ';LANGID=0x0401;CP=1256;COUNTRY=0'),
    (Value: dbSortCzech; Name: 'Czech'; Setting: ';LANGID=0x0405;CP=1250;COUNTRY=0'),
    (Value: dbSortDutch; Name: 'Dutch'; Setting: ';LANGID=0x0413;CP=1252;COUNTRY=0'),
    (Value: dbSortGreek; Name: 'Greek'; Setting: ';LANGID=0x0408;CP=1253;COUNTRY=0'),
    (Value: dbSortHebrew; Name: 'Hebrew'; Setting: ';LANGID=0x040D;CP=1255;COUNTRY=0'),
    (Value: dbSortHungarian; Name: 'Hungarian'; Setting: ';LANGID=0x040E;CP=1250;COUNTRY=0'),
    (Value: dbSortIcelandic; Name: 'Icelandic'; Setting: ';LANGID=0x040F;CP=1252;COUNTRY=0'),
    (Value: dbSortJapanese; Name: 'Japanese'; Setting: ';LANGID=0x0411;CP=932;COUNTRY=0'),
    (Value: dbSortKorean; Name: 'Korean'; Setting: ';LANGID=0x0412;CP=949;COUNTRY=0'),
    //(Value: dbSortNordic; Name: 'Nordic'; Setting: ';LANGID=0x041D;CP=1252;COUNTRY=0'),
    (Value: dbSortNorwDan; Name: 'Norwegian and Danish'; Setting: ';LANGID=0x0406;CP=1252;COUNTRY=0'),
    (Value: dbSortPolish; Name: 'Polish'; Setting: ';LANGID=0x0415;CP=1250;COUNTRY=0'),
    (Value: dbSortCyrillic; Name: 'Russian'; Setting: ';LANGID=0x0419;CP=1251;COUNTRY=0'),
    (Value: dbSortChineseSimplified; Name: 'Simplified Chinese'; Setting: ';LANGID=0x0804;CP=936;COUNTRY=0'),
    (Value: dbSortSlovenian; Name: 'Slovenian'; Setting: ';LANGID=0x0424;CP=1250;COUNTRY=0'),
    (Value: dbSortSpanish; Name: 'Spanish'; Setting: ';LANGID=0x040A;CP=1252;COUNTRY=0'),
    (Value: dbSortSwedFin; Name: 'Swedish and Finnish'; Setting: ';LANGID=0x041D;CP=1252;COUNTRY=0'),
    (Value: dbSortThai; Name: 'Thai'; Setting: ';LANGID=0x041E;CP=874;COUNTRY=0'),
    (Value: dbSortChineseTraditional; Name: 'Traditional Chinese'; Setting: ';LANGID=0x0404;CP=950;COUNTRY=0'),
    (Value: dbSortTurkish; Name: 'Turkish'; Setting: ';LANGID=0x041F;CP=1254;COUNTRY=0')
    );

  // Relations
  {
  DBRelationUnique = 1;
  DBRelationDontEnforce = 2;
  DBRelationUpdateCascade = 256;
  DBRelationDeleteCascade = 4096;
  }

  // Field types
  DBFieldTypeText = 10;
  DBFieldTypeMemo = 12;
  DBFieldAutoIncAttr = 17;

  DBFieldTypeArrSize = 14;
  DBFieldTypeArr: array [1 .. DBFieldTypeArrSize] of TDBFieldTypeRec = (
    (Value: dbText; Name: 'Text'; Desc: 'Text data (variable width)'),
    (Value: dbMemo; Name: 'Memo'; Desc: 'Memo data (extended text)'),
    // (Value: dbChar; Name: 'Char'; Desc: 'Text data (fixed width)'),
    (Value: dbBoolean; Name: 'Boolean'; Desc: 'Boolean (True/False) data'),
    (Value: dbByte; Name: 'Byte'; Desc: 'Byte (8-bit) data'),
    (Value: dbInteger; Name: 'Integer'; Desc: 'Integer data'),
    (Value: dbBigInt; Name: 'BigInt'; Desc: 'Big Integer data'),
    (Value: dbLong; Name: 'Long'; Desc: 'Long Integer data'),
    (Value: dbCurrency; Name: 'Currency'; Desc: 'Currency data'),
    // (Value: dbNumeric; Name: 'Numeric'; Desc: 'Numeric data (ODBCDirect only)'),
    // (Value: dbDecimal; Name: 'Decimal'; Desc: 'Decimal data (ODBCDirect only)'),
    (Value: dbSingle; Name: 'Single'; Desc: 'Single-precision floating-point data'),
    (Value: dbDouble; Name: 'Double'; Desc: 'Double-precision floating-point data'),
    // (Value: dbFloat; Name: 'Float'; Desc: 'Floating-point data (ODBCDirect only)'),
    (Value: dbDate; Name: 'Date'; Desc: 'Date value data'),
    // (Value: dbTime; Name: 'Time'; Desc: 'Data in time format (ODBCDirect only)'),
    // (Value: dbTimeStamp; Name: 'TimeStamp'; Desc: 'Data in time and date format (ODBCDirect only)'),
    (Value: dbBinary; Name: 'Binary'; Desc: 'Binary data'),
    (Value: dbLongBinary; Name: 'LongBinary'; Desc: 'Binary data (bitmap)'),
    (Value: dbGUID; Name: 'GUID'; Desc: 'GUID data')
    // (Value: dbVarBinary; Name: 'VarBinary'; Desc: 'Variable Binary data (ODBCDirect only)')
    // (Value: dbAttachment; Name: 'Attachment'; Desc: 'Attachment data'),
    // (Value: 102; Name: 'ComplexByte'; Desc: 'Multi-valued byte data'),
    // (Value: 103; Name: 'ComplexInteger'; Desc: 'Multi-value integer data'),
    // (Value: 104; Name: 'ComplexLong'; Desc: 'Multi-value long integer data'),
    // (Value: 105; Name: 'ComplexSingle'; Desc: 'Multi-value single-precision floating-point data'),
    // (Value: 106; Name: 'ComplexDouble'; Desc: 'Multi-value double-precision floating-point data'),
    // (Value: 107; Name: 'ComplexGUID'; Desc: 'Multi-value GUID data'),
    // (Value: 108; Name: 'ComplexDecimal'; Desc: 'Multi-value decimal data'),
    // (Value: 109; Name: 'ComplexText'; Desc: 'Multi-value Text data (variable width)')
    );

{$IFDEF DX100P}
  DBFieldROTypeArrSize = 16;
{$ELSE}
  DBFieldROTypeArrSize = 7;
{$ENDIF}
  DBFieldROTypeArr: array [1 .. DBFieldROTypeArrSize] of TDBFieldTypeRec = (
  (Value: dbChar; Name: 'Char'; Desc: 'Text data (fixed width)'),
    (Value: dbNumeric; Name: 'Numeric'; Desc: 'Numeric data (ODBCDirect only)'),
    (Value: dbDecimal;     Name: 'Decimal'; Desc: 'Decimal data (ODBCDirect only)'),
    (Value: dbFloat; Name: 'Float';  Desc: 'Floating-point data (ODBCDirect only)'),
    (Value: dbTime; Name: 'Time';    Desc: 'Data in time format (ODBCDirect only)'),
    (Value: dbTimeStamp; Name: 'TimeStamp';    Desc: 'Data in time and date format (ODBCDirect only)'),
    (Value: dbVarBinary; Name: 'VarBinary';    Desc: 'Variable Binary data (ODBCDirect only)')
{$IFDEF DX100P} ,
    (Value: dbAttachment; Name: 'Attachment'; Desc: 'Attachment data'),
    (Value: dbComplexByte; Name: 'ComplexByte'; Desc: 'Multi-valued byte data'),
    (Value: dbComplexInteger; Name: 'ComplexInteger';     Desc: 'Multi-value integer data'),
    (Value: dbComplexLong; Name: 'ComplexLong'; Desc: 'Multi-value long integer data'),
    (Value: dbComplexSingle; Name: 'ComplexSingle'; Desc: 'Multi-value single-precision floating-point data'),
    (Value: dbComplexDouble; Name: 'ComplexDouble'; Desc: 'Multi-value double-precision floating-point data'),
    (Value: dbComplexGUID; Name: 'ComplexGUID';  Desc: 'Multi-value GUID data'),
    (Value: dbComplexDecimal; Name: 'ComplexDecimal'; Desc: 'Multi-value decimal data'),
    (Value: dbComplexText; Name: 'ComplexText'; Desc: 'Multi-value Text data (variable width)')
{$ELSE}
{$ENDIF}
);

type
  TFieldRec = record
    FieldName: string;
    FieldType: integer;
    FieldSize: integer;
    Attributes: integer;
    Required: boolean;
    AllowZeroLength: boolean;
    DefaultValue: OleVariant;
    ForeignName: string;
    ValidationText: string;
    ValidationRule: string;
    //Collation: integer;
  end;

  TFieldRecArray = array of TFieldRec;

  TIndexRec = record
    IndexName: string;
    Primary, Unique, Required: boolean;
    FieldNames: string;
    Relation: boolean;
  end;

  TIndexRecArray = array of TIndexRec;

  TQueryDef = record
    Name: string;
    SQL: string;
  end;

  TQueryDefArray = array of TQueryDef;

  TRelationRec = record
    Name, Table, ForeignTable: string;
    Attributes: integer;
    Fields: TFieldRecArray;
    FieldNames, ForeignFieldNames: string;
  end;

  TRelationRecArray = array of TRelationRec;

// DBVersion
function DBVersionValueToName(Value: integer): string;
function DBVersionValueToExtension(Value: integer): string;
function DBVersionNameToValue(Name: string): integer;
function DBVersionGetList(L: TStrings): integer;

// DBLocale
function DBLocaleValueToName(Value: integer): string;
function DBLocaleValueToSetting(Value: integer): string;
function DBLocaleNameToValue(Name: string): integer;
function DBLocaleGetList(L: TStrings): integer;

// FieldType
function DBFieldTypeValueToName(Value: integer): string;
function DBFieldTypeNameToValue(Name: string): integer;
function DBFieldTypeGetList(L: TStrings): integer;

// Aux old
function IsAccess97(DBName: string): boolean;
function ReadMDBUsers(DBName: string; List: TStrings): boolean;
procedure RegDAO350;

// databases
function CreateDB(DBName: string; Locale: integer = dbSortGeneral; DBVersion: integer = DBVersionDefault): boolean;
function OpenDB(DBName, Pwd: string; OpenExclusive: boolean = true): Database;
procedure CompactDBEx(SrcName, DstName, OldPwd, NewPwd: string; OldLocale: integer =0; NewLocale: integer = 0; NewDBVersion: integer = 0);
function CompactDB(DBName, Pwd: string; NewPwd: string = ''; NewLocale: integer = 0; NewDBVersion: integer = 0): boolean;
function GetDBVersion(db: Database): integer;

// tables
procedure GetTableList(DBName, Pwd: string; List: TStrings); overload;
procedure GetTableList(db: Database; List: TStrings); overload;
function CreateTable(db: Database; TableName: string; FieldRecArray: TFieldRecArray; IndexRecArray: TIndexRecArray)
  : TableDef;
procedure RenameTable(db: Database; SrcTableName, DstTableName: string);
procedure DropTable(db: Database; TableName: string);

// table fields
function GetTableFields(DBName, Pwd, TableName: string; List: TStrings): TFieldRecArray; overload;
function GetTableFields(db: Database; TableName: string; List: TStrings): TFieldRecArray; overload;

// zable indexes
function GetTableIndexes(DBName, Pwd, TableName: string; NotRelations: boolean; List: TStrings)
  : TIndexRecArray; overload;
function GetTableIndexes(db: Database; TableName: string; NotRelations: boolean; List: TStrings)
  : TIndexRecArray; overload;
procedure DeleteIndexes(db: Database; TableName: string; LeftPrimaryKey: boolean);

// queries
function GetQueryDefs(DBName, Pwd: string; List: TStrings): TQueryDefArray; overload;
function GetQueryDefs(db: Database; List: TStrings): TQueryDefArray; overload;
procedure ClearQueryDefs(db: Database);
procedure CreateQueryDefs(db: Database; QDefArr: TQueryDefArray);
function UpdateQueryDef(db: Database; QDef: TQueryDef; NewQueryName: string = ''): QueryDef;
procedure DeleteQueryDef(db: Database; QueryName: string);

// relations
function GetRelations(DBName, Pwd: string; List: TStrings; SkipSYSRel: boolean = false): TRelationRecArray; overload;
function GetRelations(db: Database; List: TStrings; SkipSYSRel: boolean = false): TRelationRecArray; overload;
procedure CreateRelations(db: Database; RelationRecArray: TRelationRecArray);
procedure ClearRelations(db: Database);


implementation

uses rtool, rstring, rdlg;

function DBVersionValueToName(Value: integer): string;
var
  I: integer;
begin
  Result:='';
  for I:=1 to DBVersionArrSize do
    if DBVersionArr[I].Value = Value then
    begin
      Result:=DBVersionArr[I].Name;
      Exit;
    end;
end;

function DBVersionValueToExtension(Value: integer): string;
var
  I: integer;
begin
  Result:='';
  for I:=1 to DBVersionArrSize do
    if DBVersionArr[I].Value = Value then
    begin
      Result:=DBVersionArr[I].Extension;
      Exit;
    end;
end;

function DBVersionNameToValue(Name: string): integer;
var
  I: integer;
begin
  Result:=-1;
  for I:=1 to DBVersionArrSize do
    if SameText(DBVersionArr[I].Name, Name) then
    begin
      Result:=DBVersionArr[I].Value;
      Exit;
    end;
end;

function DBVersionGetList(L: TStrings): integer;
var
  I: integer;
begin
  L.Clear;
  for I:=1 to DBVersionArrSize do
    L.Add(DBVersionArr[I].Name);
  Result:=L.Count;
end;

function DBLocaleValueToName(Value: integer): string;
var
  I: integer;
begin
  Result:='';
  for I:=1 to DBLocaleArrSize do
    if DBLocaleArr[I].Value = Value then
    begin
      Result:=DBLocaleArr[I].Name;
      Exit;
    end;
end;

function DBLocaleValueToSetting(Value: integer): string;
var
  I: integer;
begin
  Result:='';
  for I:=1 to DBLocaleArrSize do
    if DBLocaleArr[I].Value = Value then
    begin
      Result:=DBLocaleArr[I].Setting;
      Exit;
    end;
end;

function DBLocaleNameToValue(Name: string): integer;
var
  I: integer;
begin
  Result:=-1;
  for I:=1 to DBLocaleArrSize do
    if SameText(DBLocaleArr[I].Name, Name) then
    begin
      Result:=DBLocaleArr[I].Value;
      Exit;
    end;
end;

function DBLocaleGetList(L: TStrings): integer;
var
  I: integer;
begin
  L.Clear;
  for I:=1 to DBLocaleArrSize do
    L.Add(DBLocaleArr[I].Name);
  Result:=L.Count;
end;

function DBFieldTypeValueToName(Value: integer): string;
var
  I: integer;
begin
  Result:='';
  for I:=1 to DBFieldTypeArrSize do
    if DBFieldTypeArr[I].Value = Value then
    begin
      Result:=DBFieldTypeArr[I].Name;
      Exit;
    end;
  for I:=1 to DBFieldROTypeArrSize do
    if DBFieldROTypeArr[I].Value = Value then
    begin
      Result:=DBFieldROTypeArr[I].Name;
      Exit;
    end;
end;

function DBFieldTypeNameToValue(Name: string): integer;
var
  I: integer;
begin
  Result:=-1;
  for I:=1 to DBFieldTypeArrSize do
    if SameText(DBFieldTypeArr[I].Name, Name) then
    begin
      Result:=DBFieldTypeArr[I].Value;
      Exit;
    end;
end;

function DBFieldTypeGetList(L: TStrings): integer;
var
  I: integer;
begin
  L.Clear;
  for I:=1 to DBFieldTypeArrSize do
    L.Add(DBFieldTypeArr[I].Name);
  Result:=L.Count;
end;

// Aux old
function IsAccess97(DBName: string): boolean;
var
  fi: file of byte;
  I: integer;
  b: byte;
begin
  Result:=false;
  AssignFile(fi, DBName);
  Reset(fi);
  try
    I:=0;
    try
      repeat
        Read(fi, b);
        Inc(I);
        if I = $15 then
        begin
          Result:=b <> 1;
          Break;
        end;
      until Eof(fi);
    except
      raise Exception.Create(sReadDataErr + #13 + Format(sFile, [DBName]));
    end;
  finally
    CloseFile(fi);
  end;
end;

procedure RegDAO350;
var
  s, p, dll, reg: string;
begin
  p:=GetAppDir;
  dll:=p + 'DAO350.dll';
  reg:=p + 'DAO350.reg';
  if not FileExists(dll) then
    raise Exception.Create(Format(sFileNotFound, [dll]));
  if not FileExists(reg) then
    raise Exception.Create(Format(sFileNotFound, [reg]));

  ShowInfo(sInitialize, nil);
  s:=Format('/u /s "%s"', [dll]);
  RunFile('Regsvr32.exe', s, '', SW_SHOW);
  Sleep(1000);
  s:=Format('/s "%s"', [dll]);
  RunFile('Regsvr32.exe', s, '', SW_SHOW);
  Sleep(1000);
  s:=Format('/s "%s"', [reg]);
  RunFile('Regedit.exe', s, '', SW_SHOW);
  Sleep(1000);
  ShowInfo('', nil);
end;

function ReadMDBUsers(DBName: string; List: TStrings): boolean;
const
  DLLFileName = 'MSLDBUSR.dll';
type
  TLDBUser_GetUsers = function(var UserBuffer: PSafeArray; DatabaseName: PChar; Options: integer): integer; stdcall;
  TLDBUser_GetError = function(ErrorNo: integer): PChar; stdcall;
var
  DLLHandle: THandle;
  LDBUser_GetUsers: TLDBUser_GetUsers;
  LDBUser_GetError: TLDBUser_GetError;
  UserCount: integer;
  UserBuffer: Variant;
  Options: integer;
  I: integer;
begin
  Result:=false;
  List.Clear;
  WCursor;
  try
    DLLHandle:=LoadLibrary(DLLFileName);
    if DLLHandle <= 0 then
      List.Add(Format(sDLLNotFound, [DLLFileName]))
    else
      try
        @LDBUser_GetUsers:=GetProcAddress(DLLHandle, 'LDBUser_GetUsers');
        @LDBUser_GetError:=GetProcAddress(DLLHandle, 'LDBUser_GetError');
        if (@LDBUser_GetUsers = nil) or (@LDBUser_GetError = nil) then
          List.Add(Format(sDLLNotFound, [DLLFileName]))
        else
        begin
          UserBuffer:=VarArrayCreate([1, 1], varOleStr);
          // options: 1=all users 2=connected users
          // 4=user causing corrupted DB; 8=count of users
          Options:=2;
          UserCount:=LDBUser_GetUsers(PSafeArray(TVarData(UserBuffer).VArray), PChar(DBName), Options);
          if UserCount < 0 then
            List.Add(LDBUser_GetError(UserCount))
          else
          begin
            for I:=1 to UserCount do
              List.Add(PChar(TVarData(UserBuffer[I]).VPointer));
            Result:=true;
          end;
        end;
      finally
        FreeLibrary(DLLHandle);
      end;
  finally
    NCursor;
  end;
end;

// ============================================================================
// databases

function CreateDB(DBName: string; Locale: integer = dbSortGeneral; DBVersion: integer = DBVersionDefault): boolean;
var
  dbEngine: _DBEngine;
  LocaleStr: string;
begin
  LocaleStr:=DBLocaleValueToSetting(Locale);
  dbEngine:=CoDBEngine.Create;
  dbEngine.CreateDatabase(DBName, LocaleStr, DBVersion);
  dbEngine:=nil;
  Result:=true;
end;

function OpenDB(DBName, Pwd: string; OpenExclusive: boolean = true): Database;
var
  db: Database;
  dbEngine: _DBEngine;
begin
  if Pwd <> '' then
    Pwd:=';pwd=' + Pwd;
  dbEngine:=CoDBEngine.Create;
  db:=dbEngine.OpenDatabase(DBName, OpenExclusive, false, Pwd);
  Result:=db;
end;

procedure CompactDBEx(SrcName, DstName, OldPwd, NewPwd: string; OldLocale: integer =0; NewLocale: integer = 0; NewDBVersion: integer = 0);
var
  dbEngine: _DBEngine;
begin
  if OldPwd <> '' then
    OldPwd:=';pwd=' + OldPwd;
  if NewPwd=#0 then
    NewPwd:=';pwd='
  else
  if NewPwd <> '' then
    NewPwd:=';pwd=' + NewPwd;
  dbEngine:=CoDBEngine.Create;
  dbEngine.CompactDatabase(SrcName, DstName, DBLocaleValueToSetting(NewLocale)+NewPwd, NewDBVersion, DBLocaleValueToSetting(OldLocale)+OldPwd);
end;

function CompactDB(DBName, Pwd: string; NewPwd: string = ''; NewLocale: integer = 0; NewDBVersion: integer = 0): boolean;
var
  Dest, Old: string;
  procedure DelFile(f: string);
  begin
    if FileExists(f) then
      if not DeleteFile(f) then
        raise Exception.Create('Cannot delete file: ' + f);
  end;

begin
  Result:=false;
  try
    Dest:=ExtractFilePath(DBName) + '_compact_.mdb';
    Old:=ExtractFilePath(DBName) + '_old_.mdb';
    DelFile(Dest);
    DelFile(Old);
    if NewPwd = '' then
      NewPwd:=Pwd;
    CompactDBEx(DBName, Dest, Pwd, NewPwd, 0, NewLocale, NewDBVersion);
    if not RenameFile(DBName, Old) then
      raise Exception.Create('Cannot rename file: ' + DBName);
    if not RenameFile(Dest, DBName) then
      raise Exception.Create('Cannot rename file: ' + Dest);
    DelFile(Dest);
    DelFile(Old);
    Result:=true;
  except
    on E: Exception do
      ShowErr(sCompressDataErr, E, false);
  end;
  NCursor;
end;

function GetDBVersion(db: Database): integer;
var i: integer;
 ver: string;
begin
  Result:=-1;
  ver:=db.Version;
  for I:=1 to DBVersionArrSize do
    if Pos(ver, DBVersionArr[I].Name)>0 then
    begin
      Result:=DBVersionArr[I].Value;
      Exit;
    end;
end;

// ============================================================================
// Tables

procedure GetTableList(DBName, Pwd: string; List: TStrings);
var
  db: Database;
begin
  db:=OpenDB(DBName, Pwd);
  GetTableList(db, List);
end;

procedure GetTableList(db: Database; List: TStrings);
var
  I: integer;
  s: string;
begin
  List.Clear;
  for I:=0 to db.TableDefs.Count - 1 do
  begin
    s:=db.TableDefs[I].Name;
    if AnsiPosEx('MSYS', s, 1, false) <> 1 then
      List.Add(s);
  end;
end;

function CreateTable(db: Database; TableName: string; FieldRecArray: TFieldRecArray; IndexRecArray: TIndexRecArray)
  : TableDef;
var
  TblDef: TableDef;
  FldDef: Field;
  IdxDef: Index;
  I: integer;
  sIdxFldNames, sIdxFld: string;
begin
  TblDef:=db.CreateTableDef(TableName, 0, '', '');
  for I:=0 to High(FieldRecArray) do
  begin
    FldDef:=TblDef.CreateField(FieldRecArray[I].FieldName, FieldRecArray[I].FieldType, FieldRecArray[I].FieldSize);
    FldDef.Set_Attributes(FieldRecArray[I].Attributes);
    FldDef.Set_Required(FieldRecArray[I].Required);
    FldDef.Set_DefaultValue(FieldRecArray[I].DefaultValue);
    // text file?
    if FieldRecArray[I].FieldType = 10 then
      FldDef.Set_AllowZeroLength(FieldRecArray[I].AllowZeroLength);
    FldDef.Set_ValidationText(FieldRecArray[I].ValidationText);
    FldDef.Set_ValidationRule(FieldRecArray[I].ValidationRule);
    TblDef.Fields.Append(FldDef);
  end;
  for I:=0 to High(IndexRecArray) do
  begin
    IdxDef:=TblDef.CreateIndex(IndexRecArray[I].IndexName);
    sIdxFldNames:=IndexRecArray[I].FieldNames;
    repeat
      sIdxFld:=Trim(rtool.RemoveStrPart(sIdxFldNames, ';', true));
      if sIdxFld = '' then
        Break;
      FldDef:=IdxDef.CreateField(sIdxFld, EmptyParam, EmptyParam);
      IdxDef.Fields.Append(FldDef);
    until false;
    IdxDef.Set_Unique(IndexRecArray[I].Unique);
    IdxDef.Set_Required(IndexRecArray[I].Required);
    // set primary as last
    IdxDef.Set_Primary(IndexRecArray[I].Primary);
    TblDef.Indexes.Append(IdxDef);
  end;
  db.TableDefs.Append(TblDef);
  db.TableDefs.Refresh;
  Result:=TblDef;
end;

procedure DropTable(db: Database; TableName: string);
begin
  db.TableDefs.Delete(TableName);
  db.TableDefs.Refresh;
end;

procedure RenameTable(db: Database; SrcTableName, DstTableName: string);
var
  tb: TableDef;
begin
  tb:=db.TableDefs[SrcTableName];
  tb.Set_Name(DstTableName);
  db.TableDefs.Refresh;
end;

// ============================================================================
// Table Fields

function GetTableFields(DBName, Pwd, TableName: string; List: TStrings): TFieldRecArray;
var
  db: Database;
begin
  db:=OpenDB(DBName, Pwd);
  Result:=GetTableFields(db, TableName, List);
end;

function GetTableFields(db: Database; TableName: string; List: TStrings): TFieldRecArray;
var
  TblDef: TableDef;
  I: integer;
begin
  Result:=nil;
  if Assigned(List) then
    List.Clear;
  TblDef:=db.TableDefs[TableName];
  if not Assigned(TblDef) then
    Exit;
  SetLength(Result, TblDef.Fields.Count);
  for I:=0 to TblDef.Fields.Count - 1 do
  begin
    Result[I].FieldName:=TblDef.Fields[I].Name;
    Result[I].FieldType:=TblDef.Fields[I].Type_;
    Result[I].FieldSize:=TblDef.Fields[I].Size;
    Result[I].Attributes:=TblDef.Fields[I].Attributes;
    Result[I].Required:=TblDef.Fields[I].Required;
    Result[I].AllowZeroLength:=TblDef.Fields[I].AllowZeroLength;
    Result[I].DefaultValue:=TblDef.Fields[I].DefaultValue;
    Result[I].ValidationText:=TblDef.Fields[I].ValidationText;
    Result[I].ValidationRule:=TblDef.Fields[I].ValidationRule;
    //Result[I].Collation:=TblDef.Fields[I].CollatingOrder;
    if Assigned(List) then
      List.Add(TblDef.Fields[I].Name);
  end;
end;

// ============================================================================
// Table Indexes

function GetTableIndexes(DBName, Pwd, TableName: string; NotRelations: boolean; List: TStrings): TIndexRecArray;
var
  db: Database;
begin
  db:=OpenDB(DBName, Pwd);
  Result:=GetTableIndexes(db, TableName, NotRelations, List);
end;

function GetTableIndexes(db: Database; TableName: string; NotRelations: boolean; List: TStrings): TIndexRecArray;
var
  TblDef: TableDef;
  I, p: integer;
  sFld: string;
  RelList: TStrings;
begin
  Result:=nil;
  if Assigned(List) then
    List.Clear;
  TblDef:=db.TableDefs[TableName];
  if not Assigned(TblDef) then
    Exit;
  SetLength(Result, TblDef.Indexes.Count);
  p:=0;
  RelList:=TStringList.Create;
  try
    GetRelations(db, RelList);
    for I:=0 to TblDef.Indexes.Count - 1 do
    begin
      if (not NotRelations) or (RelList.IndexOf(TblDef.Indexes[I].Name) < 0) then
      begin
        // ??? fields is list in format +FieldName;+FieldName ...
        sFld:=ClearStr('+', TblDef.Indexes[I].Fields);
        Result[p].FieldNames:=sFld;
        Result[p].IndexName:=TblDef.Indexes[I].Name;
        Result[p].Primary:=TblDef.Indexes[I].Primary;
        Result[p].Unique:=TblDef.Indexes[I].Unique;
        Result[p].Required:=TblDef.Indexes[I].Required;
        Result[p].Relation:=RelList.IndexOf(TblDef.Indexes[I].Name) >= 0;
        if Assigned(List) then
          List.Add(TblDef.Indexes[I].Name);
        Inc(p);
      end;
    end;
    SetLength(Result, p);
  finally
    RelList.Clear;
  end;
end;

procedure DeleteIndexes(db: Database; TableName: string; LeftPrimaryKey: boolean);
var
  TblDef: TableDef;
  I: integer;
begin
  TblDef:=db.TableDefs[TableName];
  for I:=TblDef.Indexes.Count - 1 downto 0 do
  begin
    if not(LeftPrimaryKey and TblDef.Indexes[I].Primary) then
      try
        TblDef.Indexes.Delete(TblDef.Indexes[I].Name);
      except
      end;
  end;
  db.TableDefs.Refresh;
end;

// ============================================================================
// Queries

function GetQueryDefs(DBName, Pwd: string; List: TStrings): TQueryDefArray;
var
  db: Database;
begin
  db:=OpenDB(DBName, Pwd);
  Result:=GetQueryDefs(db, List);
end;

function GetQueryDefs(db: Database; List: TStrings): TQueryDefArray;
var
  I: integer;
begin
  if Assigned(List) then
    List.Clear;
  SetLength(Result, db.QueryDefs.Count);
  for I:=0 to db.QueryDefs.Count - 1 do
  begin
    Result[I].Name:=db.QueryDefs[I].Name;
    Result[I].SQL:=db.QueryDefs[I].SQL;
    if Assigned(List) then
      List.Add(db.QueryDefs[I].Name);
  end;
end;

procedure ClearQueryDefs(db: Database);
var
  I: integer;
begin
  for I:=db.QueryDefs.Count - 1 downto 0 do
    db.QueryDefs.Delete(db.QueryDefs[I].Name);
  db.QueryDefs.Refresh;
end;

procedure CreateQueryDefs(db: Database; QDefArr: TQueryDefArray);
var
  I: integer;
begin
  for I:=0 to High(QDefArr) do
    db.CreateQueryDef(QDefArr[I].Name, QDefArr[I].SQL);
  db.QueryDefs.Refresh;
end;

function UpdateQueryDef(db: Database; QDef: TQueryDef; NewQueryName: string = ''): QueryDef;
var
  I: integer;
begin
  for I:=db.QueryDefs.Count - 1 downto 0 do
    if SameText(db.QueryDefs[I].Name, QDef.Name) then
      db.QueryDefs.Delete(QDef.Name);
  db.QueryDefs.Refresh;
  if NewQueryName='' then
    NewQueryName:=QDef.Name;
  Result:=db.CreateQueryDef(NewQueryName, QDef.SQL);
  db.QueryDefs.Refresh;
end;

procedure DeleteQueryDef(db: Database; QueryName: string);
var
  I: integer;
begin
  for I:=db.QueryDefs.Count - 1 downto 0 do
    if SameText(db.QueryDefs[I].Name, QueryName) then
      db.QueryDefs.Delete(QueryName);
  db.QueryDefs.Refresh;
end;

// ============================================================================
// Relations

function GetRelations(DBName, Pwd: string; List: TStrings; SkipSYSRel: boolean = false): TRelationRecArray;
var
  db: Database;
begin
  db:=OpenDB(DBName, Pwd);
  Result:=GetRelations(db, List, SkipSYSRel);
end;

function GetRelations(db: Database; List: TStrings; SkipSYSRel: boolean = false): TRelationRecArray;
  procedure AddToList(var List: string; Item: string);
  begin
    if List <> '' then
      List:=List + ';';
    List:=List + Item;
  end;

var
  RelDef: Relation;
  I, j, p: integer;
begin
  if Assigned(List) then
    List.Clear;
  SetLength(Result, db.Relations.Count);
  p:=0;
  for I:=0 to db.Relations.Count - 1 do
  begin
    RelDef:=db.Relations[I];
    if SkipSYSRel and (AnsiPosEx('MSYS', RelDef.Name, 1, false) = 1) then
      Continue;
    Result[p].Name:=RelDef.Name;
    Result[p].Table:=RelDef.Table;
    Result[p].ForeignTable:=RelDef.ForeignTable;
    Result[p].Attributes:=RelDef.Attributes;
    SetLength(Result[p].Fields, RelDef.Fields.Count);
    Result[p].FieldNames:='';
    Result[p].ForeignFieldNames:='';
    for j:=0 to RelDef.Fields.Count - 1 do
    begin
      Result[p].Fields[j].FieldName:=RelDef.Fields[j].Name;
      Result[p].Fields[j].ForeignName:=RelDef.Fields[j].ForeignName;
      AddToList(Result[p].FieldNames, Result[p].Fields[j].FieldName);
      AddToList(Result[p].ForeignFieldNames, Result[p].Fields[j].ForeignName);
    end;
    if Assigned(List) then
      List.Add(RelDef.Name);
    Inc(p);
    // if UseDemoRecLimit and (p>=DemoMaxRec) then Break;
  end;
  SetLength(Result, p);
end;

procedure CreateRelations(db: Database; RelationRecArray: TRelationRecArray);
var
  RelDef: Relation;
  FldDef: Field;
  I, j: integer;
begin
  for I:=0 to High(RelationRecArray) do
  begin
    RelDef:=db.CreateRelation(RelationRecArray[I].Name, RelationRecArray[I].Table, RelationRecArray[I].ForeignTable,
      RelationRecArray[I].Attributes);
    for j:=0 to High(RelationRecArray[I].Fields) do
    begin
      FldDef:=RelDef.CreateField(RelationRecArray[I].Fields[j].FieldName, EmptyParam, EmptyParam);
      FldDef.Set_ForeignName(RelationRecArray[I].Fields[j].ForeignName);
      RelDef.Fields.Append(FldDef);
    end;
    db.Relations.Append(RelDef);
  end;
  db.Relations.Refresh;
end;

procedure ClearRelations(db: Database);
var
  I: integer;
begin
  for I:=db.Relations.Count - 1 downto 0 do
    db.Relations.Delete(db.Relations.Item[I].Name);
  db.Relations.Refresh;
end;

end.
